@@ -32,9 +32,9 @@ GetOptions(\%args, qw(
     os=s shell=s toolchain=s compiler=s
     ar=s cc=s ld=s make=s has-sha has-libuv
     static has-libtommath has-libatomic_ops
-    has-dyncall
+    has-dyncall has-libffi
     build=s host=s big-endian jit! enable-jit lua=s has-dynasm
-    prefix=s bindir=s libdir=s mastdir=s make-install asan),
+    prefix=s bindir=s libdir=s mastdir=s make-install asan ubsan),
     'no-optimize|nooptimize' => sub { $args{optimize} = 0 },
     'no-debug|nodebug' => sub { $args{debug} = 0 }
 ) or die "See --help for further information\n";
@@ -80,6 +80,7 @@ $args{'has-libuv'}         //= 0;
 $args{'has-libatomic_ops'} //= 0;
 $args{'has-dynasm'}        //= 0;
 $args{'asan'}              //= 0;
+$args{'ubsan'}             //= 0;
 
 # jit is default
 $args{'jit'}               //= 1;
@@ -135,6 +136,7 @@ $config{crossconf} //= '';
 $config{dllimport} //= '';
 $config{dllexport} //= '';
 $config{dlllocal}  //= '';
+$config{translate_newline_output} //= 0;
 
 # assume the compiler can be used as linker frontend
 $config{ld}           //= $config{cc};
@@ -226,13 +228,20 @@ else {
     $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/dynasm';
 }
 
-if ($args{'has-dyncall'}) {
+if ($args{'has-libffi'}) {
+    $config{nativecall_backend} = 'libffi';
+    unshift @{$config{usrlibs}}, 'ffi';
+    push @{$config{defs}}, 'HAVE_LIBFFI';
+}
+elsif ($args{'has-dyncall'}) {
     unshift @{$config{usrlibs}}, 'dyncall_s', 'dyncallback_s', 'dynload_s';
     $defaults{-thirdparty}->{dc}  = undef;
     $defaults{-thirdparty}->{dcb} = undef;
     $defaults{-thirdparty}->{dl}  = undef;
+    $config{nativecall_backend} = 'dyncall';
 }
 else {
+    $config{nativecall_backend} = 'dyncall';
     $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/dyncall/dynload'
                         . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncall'
                         . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncallback';
@@ -264,7 +273,8 @@ $config{jit} //= '$(JIT_STUB)';
 $config{ldlibs} = join ' ',
     (map { sprintf $config{ldusr}, $_; } @{$config{usrlibs}}),
     (map { sprintf $config{ldsys}, $_; } @{$config{syslibs}});
-$config{ldlibs} .= ' -lasan' if $args{asan};
+$config{ldlibs} = ' -lasan ' . $config{ldlibs} if $args{asan};
+$config{ldlibs} = ' -lubsan ' . $config{ldlibs} if $args{ubsan};
 # macro defs
 $config{ccdefflags} = join ' ', map { $config{ccdef} . $_ } @{$config{defs}};
 
@@ -283,7 +293,9 @@ push @cflags, $config{ccinstflags}  if $args{instrument};
 push @cflags, $config{ccwarnflags};
 push @cflags, $config{ccdefflags};
 push @cflags, $config{ccshared}     unless $args{static};
-push @cflags, '-fno-omit-frame-pointer -fsanitize=address' if $args{asan};
+push @cflags, '-fno-omit-frame-pointer' if $args{asan} or $args{ubsan};
+push @cflags, '-fsanitize=address' if $args{asan};
+push @cflags, '-fsanitize=undefined' if $args{ubsan};
 push @cflags, $ENV{CFLAGS} if $ENV{CFLAGS};
 push @cflags, $ENV{CPPFLAGS} if $ENV{CPPFLAGS};
 $config{cflags} = join ' ', @cflags;
@@ -321,6 +333,8 @@ else {
     $config{libdir}    = '@prefix@/lib' if ! $args{libdir};
 }
 
+$config{mainlibs} = '-lubsan ' . $config{mainlibs} if $args{ubsan};
+
 # some toolchains generate garbage
 my @auxfiles = @{ $defaults{-auxfiles} };
 $config{auxclean} = @auxfiles ? '$(RM) ' . join ' ', @auxfiles : '@:';
@@ -340,6 +354,7 @@ else {
     build::probe::ptr_size_native(\%config, \%defaults);
 }
 
+build::probe::C_type_bool(\%config, \%defaults);
 build::probe::computed_goto(\%config, \%defaults);
 build::probe::pthread_yield(\%config, \%defaults);
 
@@ -360,7 +375,7 @@ print dots('Configuring 3rdparty libs');
 my @thirdpartylibs;
 my $thirdparty = $defaults{-thirdparty};
 
-for (keys %$thirdparty) {
+for (sort keys %$thirdparty) {
     my $current = $thirdparty->{$_};
     my @keys = ( "${_}lib", "${_}objects", "${_}rule", "${_}clean");
 
@@ -671,7 +686,7 @@ sub hardfail {
 
 sub write_backend_config {
     $config{backendconfig} = '';
-    for my $k (keys %config) {
+    for my $k (sort keys %config) {
         next if $k eq 'backendconfig';
         my $v = $config{$k};
         
@@ -708,7 +723,7 @@ __END__
                    [--static] [--prefix]
                    [--has-libtommath] [--has-sha] [--has-libuv]
                    [--has-libatomic_ops] [--has-dynasm]
-                   [--lua <lua>] [--asan] [--no-jit]
+                   [--lua <lua>] [--asan] [--ubsan] [--no-jit]
 
     ./Configure.pl --build <build-triple> --host <host-triple>
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
@@ -785,6 +800,17 @@ options.
 Build with AddressSanitizer (ASAN) support. Requires clang and LLVM 3.1 or newer.
 See L<https://code.google.com/p/address-sanitizer/wiki/AddressSanitizer>
 
+You can use C<ASAN_OPTIONS> to configure ASAN at runtime; for example, to disable
+memory leak checking (which can make Rakudo fail to build), you can set the following:
+
+    export ASAN_OPTIONS=detect_leaks=0
+
+A full list of options is displayed if you set C<ASAN_OPTIONS> to C<help=1>.
+
+=item --ubsan
+
+Build with Undefined Behaviour sanitizer support.
+
 =item --ld <ld>
 
 Explicitly set the linker without affecting other configuration
@@ -845,6 +871,8 @@ Build and install MoarVM in addition to configuring it.
 
 =item --has-dyncall
 
+=item --has-libffi
+
 =item --no-jit
 
 Disable JIT compiler, which is enabled by default to JIT-compile hot frames.
