@@ -1,5 +1,15 @@
 #include "moar.h"
 
+MVMint64 MVM_6model_container_iscont_rw(MVMThreadContext *tc, MVMObject *cont) {
+    if (cont && IS_CONCRETE(cont)) {
+        const MVMContainerSpec *cs = STABLE(cont)->container_spec;
+        if (cs && cs->can_store(tc, cont)) {
+            return 1;
+        }
+    }
+    return 0;
+}
+
 /* ***************************************************************************
  * CodePair container configuration: container with FETCH/STORE code refs
  * ***************************************************************************/
@@ -75,12 +85,7 @@ static void code_pair_gc_mark_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWor
 }
 
 static void code_pair_gc_free_data(MVMThreadContext *tc, MVMSTable *st) {
-    CodePairContData *data = (CodePairContData *)st->container_data;
-
-    if (data) {
-        MVM_free(data);
-        st->container_data = NULL;
-    }
+    MVM_free_null(st->container_data);
 }
 
 static void code_pair_serialize(MVMThreadContext *tc, MVMSTable *st, MVMSerializationWriter *writer) {
@@ -97,6 +102,10 @@ static void code_pair_deserialize(MVMThreadContext *tc, MVMSTable *st, MVMSerial
     MVM_ASSIGN_REF(tc, &(st->header), data->store_code, MVM_serialization_read_ref(tc, reader));
 }
 
+static MVMint32 code_pair_can_store(MVMThreadContext *tc, MVMObject *cont) {
+    return 1;
+}
+
 static const MVMContainerSpec code_pair_spec = {
     "code_pair",
     code_pair_fetch,
@@ -113,6 +122,7 @@ static const MVMContainerSpec code_pair_spec = {
     code_pair_gc_free_data,
     code_pair_serialize,
     code_pair_deserialize,
+    code_pair_can_store,
     0
 };
 
@@ -169,6 +179,9 @@ static void native_ref_fetch_i(MVMThreadContext *tc, MVMObject *cont, MVMRegiste
         case MVM_NATIVEREF_POSITIONAL:
             res->i64 = MVM_nativeref_read_positional_i(tc, cont);
             break;
+        case MVM_NATIVEREF_MULTIDIM:
+            res->i64 = MVM_nativeref_read_multidim_i(tc, cont);
+            break;
         default:
             MVM_exception_throw_adhoc(tc, "Unknown native int reference kind");
     }
@@ -188,6 +201,9 @@ static void native_ref_fetch_n(MVMThreadContext *tc, MVMObject *cont, MVMRegiste
         case MVM_NATIVEREF_POSITIONAL:
             res->n64 = MVM_nativeref_read_positional_n(tc, cont);
             break;
+        case MVM_NATIVEREF_MULTIDIM:
+            res->n64 = MVM_nativeref_read_multidim_n(tc, cont);
+            break;
         default:
             MVM_exception_throw_adhoc(tc, "Unknown native num reference kind");
     }
@@ -207,6 +223,9 @@ static void native_ref_fetch_s(MVMThreadContext *tc, MVMObject *cont, MVMRegiste
         case MVM_NATIVEREF_POSITIONAL:
             res->s = MVM_nativeref_read_positional_s(tc, cont);
             break;
+        case MVM_NATIVEREF_MULTIDIM:
+            res->s = MVM_nativeref_read_multidim_s(tc, cont);
+            break;
         default:
             MVM_exception_throw_adhoc(tc, "Unknown native str reference kind");
     }
@@ -250,6 +269,9 @@ static void native_ref_store_i(MVMThreadContext *tc, MVMObject *cont, MVMint64 v
         case MVM_NATIVEREF_POSITIONAL:
             MVM_nativeref_write_positional_i(tc, cont, value);
             break;
+        case MVM_NATIVEREF_MULTIDIM:
+            MVM_nativeref_write_multidim_i(tc, cont, value);
+            break;
         default:
             MVM_exception_throw_adhoc(tc, "Unknown native int reference kind");
     }
@@ -269,6 +291,9 @@ static void native_ref_store_n(MVMThreadContext *tc, MVMObject *cont, MVMnum64 v
         case MVM_NATIVEREF_POSITIONAL:
             MVM_nativeref_write_positional_n(tc, cont, value);
             break;
+        case MVM_NATIVEREF_MULTIDIM:
+            MVM_nativeref_write_multidim_n(tc, cont, value);
+            break;
         default:
             MVM_exception_throw_adhoc(tc, "Unknown native num reference kind");
     }
@@ -288,6 +313,9 @@ static void native_ref_store_s(MVMThreadContext *tc, MVMObject *cont, MVMString
         case MVM_NATIVEREF_POSITIONAL:
             MVM_nativeref_write_positional_s(tc, cont, value);
             break;
+        case MVM_NATIVEREF_MULTIDIM:
+            MVM_nativeref_write_multidim_s(tc, cont, value);
+            break;
         default:
             MVM_exception_throw_adhoc(tc, "Unknown native str reference kind");
     }
@@ -318,6 +346,10 @@ static void native_ref_deserialize(MVMThreadContext *tc, MVMSTable *st, MVMSeria
     /* Nothing to do. */
 }
 
+static MVMint32 native_ref_can_store(MVMThreadContext *tc, MVMObject *cont) {
+    return 1;
+}
+
 static const MVMContainerSpec native_ref_spec = {
     "native_ref",
     native_ref_fetch,
@@ -334,6 +366,7 @@ static const MVMContainerSpec native_ref_spec = {
     NULL, /* gc_free_data */
     native_ref_serialize,
     native_ref_deserialize,
+    native_ref_can_store,
     1
 };
 
@@ -454,6 +487,17 @@ void MVM_6model_container_decont_s(MVMThreadContext *tc, MVMObject *cont, MVMReg
         res->s = MVM_repr_get_str(tc, cont);
 }
 
+/* If it's a container, do a fetch_i. Otherwise, try to unbox the received
+ * value as a native unsigned integer. */
+void MVM_6model_container_decont_u(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
+    const MVMContainerSpec *cs = STABLE(cont)->container_spec;
+    if (cs && IS_CONCRETE(cont))
+        /* XXX We need a fetch_u at some point. */
+        cs->fetch_i(tc, cont, res);
+    else
+        res->u64 = MVM_repr_get_uint(tc, cont);
+}
+
 /* Checks we have a container, and provided we do, assigns an int into it. */
 void MVM_6model_container_assign_i(MVMThreadContext *tc, MVMObject *cont, MVMint64 value) {
     const MVMContainerSpec *cs = STABLE(cont)->container_spec;
