@@ -129,10 +129,7 @@ static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     }
 
     /* If we replaced the object body, free the replacement. */
-    if (((MVMP6opaque *)obj)->body.replaced) {
-        MVM_free(((MVMP6opaque *)obj)->body.replaced);
-        ((MVMP6opaque *)obj)->body.replaced = NULL;
-    }
+    MVM_free(((MVMP6opaque *)obj)->body.replaced);
 }
 
 /* Marks the representation data in an STable.*/
@@ -181,30 +178,31 @@ static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
     if (repr_data->name_to_index_mapping) {
         MVMP6opaqueNameMap *cur_map_entry = repr_data->name_to_index_mapping;
         while (cur_map_entry->class_key != NULL) {
-            MVM_checked_free_null(cur_map_entry->names);
-            MVM_checked_free_null(cur_map_entry->slots);
+            MVM_free(cur_map_entry->names);
+            MVM_free(cur_map_entry->slots);
             cur_map_entry++;
         }
-        MVM_checked_free_null(repr_data->name_to_index_mapping);
+        MVM_free(repr_data->name_to_index_mapping);
     }
 
-    MVM_checked_free_null(repr_data->attribute_offsets);
-    MVM_checked_free_null(repr_data->flattened_stables);
-    MVM_checked_free_null(repr_data->auto_viv_values);
-    MVM_checked_free_null(repr_data->unbox_slots);
-    MVM_checked_free_null(repr_data->gc_obj_mark_offsets);
-    MVM_checked_free_null(repr_data->initialize_slots);
-    MVM_checked_free_null(repr_data->gc_mark_slots);
-    MVM_checked_free_null(repr_data->gc_cleanup_slots);
+    MVM_free(repr_data->attribute_offsets);
+    MVM_free(repr_data->flattened_stables);
+    MVM_free(repr_data->auto_viv_values);
+    MVM_free(repr_data->unbox_slots);
+    MVM_free(repr_data->gc_obj_mark_offsets);
+    MVM_free(repr_data->initialize_slots);
+    MVM_free(repr_data->gc_mark_slots);
+    MVM_free(repr_data->gc_cleanup_slots);
 
-    MVM_checked_free_null(st->REPR_data);
+    MVM_free(st->REPR_data);
 }
 
 /* Helper for complaining about attribute access errors. */
 MVM_NO_RETURN
 static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name) {
-    MVMuint64 output_size;
-    MVM_exception_throw_adhoc(tc, "P6opaque: no such attribute '%s'", MVM_string_ascii_encode(tc, name, &output_size));
+    char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+    char *waste[] = { c_name, NULL };
+    MVM_exception_throw_adhoc_free(tc, waste, "P6opaque: no such attribute '%s'", c_name);
 }
 
 /* Gets the current value for an attribute. */
@@ -501,6 +499,36 @@ static MVMString * get_str(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
     }
 }
 
+/* Used with boxing. Sets an unsigned integer value, for representations that can hold
+ * one. */
+static void set_uint(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMuint64 value) {
+    MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
+    data = MVM_p6opaque_real_data(tc, data);
+    if (repr_data->unbox_int_slot >= 0) {
+        MVMSTable *st = repr_data->flattened_stables[repr_data->unbox_int_slot];
+        st->REPR->box_funcs.set_uint(tc, st, root, (char *)data + repr_data->attribute_offsets[repr_data->unbox_int_slot], value);
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "This type cannot box a native integer");
+    }
+}
+
+/* Used with boxing. Gets an unsigned integer value, for representations that can
+ * hold one. */
+static MVMuint64 get_uint(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data) {
+    MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
+    data = MVM_p6opaque_real_data(tc, data);
+    if (repr_data->unbox_int_slot >= 0) {
+        MVMSTable *st = repr_data->flattened_stables[repr_data->unbox_int_slot];
+        return st->REPR->box_funcs.get_uint(tc, st, root, (char *)data + repr_data->attribute_offsets[repr_data->unbox_int_slot]);
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "This type cannot unbox to a native integer");
+    }
+}
+
 static void * get_boxed_ref(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMuint32 repr_id) {
     MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
     data = MVM_p6opaque_real_data(tc, data);
@@ -581,7 +609,6 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
      * attributes. */
     mro_count   = REPR(info)->elems(tc, STABLE(info), info, OBJECT_BODY(info));
     mro_pos     = mro_count;
-    num_attrs   = 0;
     total_attrs = 0;
     while (mro_pos--) {
         /* Get info for the class at the current position. */
@@ -1506,13 +1533,14 @@ static const MVMREPROps this_repr = {
         get_num,
         set_str,
         get_str,
+        set_uint,
+        get_uint,
         get_boxed_ref
     },    /* box_funcs */
     {
         at_pos,
         bind_pos,
         set_elems,
-        NULL,
         push,
         pop,
         unshift,
