@@ -240,14 +240,14 @@ void MVM_coerce_smart_stringify(MVMThreadContext *tc, MVMObject *obj, MVMRegiste
 }
 
 MVMint64 MVM_coerce_s_i(MVMThreadContext *tc, MVMString *s) {
-    char     *enc = MVM_string_ascii_encode(tc, s, NULL);
+    char     *enc = MVM_string_ascii_encode(tc, s, NULL, 0);
     MVMint64  i   = strtoll(enc, NULL, 10);
     MVM_free(enc);
     return i;
 }
 
 MVMnum64 MVM_coerce_s_n(MVMThreadContext *tc, MVMString *s) {
-    char     *enc = MVM_string_ascii_encode(tc, s, NULL);
+    char     *enc = MVM_string_ascii_encode(tc, s, NULL, 0);
     MVMnum64  n;
     if (strcmp(enc, "NaN") == 0)
         n = MVM_num_nan(tc);
@@ -363,7 +363,7 @@ MVMObject * MVM_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *str, MVMi
         else if (ch >= 'A' && ch <= 'Z') ch = ch - 'A' + 10;
         else if (ch >= 0xFF21 && ch <= 0xFF3A) ch = ch - 0xFF21 + 10; /* uppercase fullwidth */
         else if (ch >= 0xFF41 && ch <= 0xFF5A) ch = ch - 0xFF41 + 10; /* lowercase fullwidth */
-        else if (MVM_unicode_codepoint_has_property_value(tc, ch, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY, 
+        else if (ch > 0 && MVM_unicode_codepoint_has_property_value(tc, ch, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY, 
                 MVM_unicode_cname_to_property_value_code(tc, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY, STR_WITH_LEN("Nd")))) {
             /* As of Unicode 6.0.0, we know that Nd category numerals are within
              * the range 0..9
@@ -445,3 +445,12 @@ void MVM_box_str(MVMThreadContext *tc, MVMString *value, MVMObject *type,
             dst->o = box;
         });
 }
+
+void MVM_box_uint(MVMThreadContext *tc, MVMuint64 value, MVMObject *type,
+             MVMRegister * dst) {
+    MVMObject *box = REPR(type)->allocate(tc, STABLE(type));
+    if (REPR(box)->initialize)
+        REPR(box)->initialize(tc, STABLE(box), box, OBJECT_BODY(box));
+    REPR(box)->box_funcs.set_uint(tc, STABLE(box), box, OBJECT_BODY(box), value);
+    dst->o = box;
+}
