@@ -47,6 +47,7 @@ typedef struct {
     MVMCallsite      *cur_call;
     MVMuint16         cur_arg;
     MVMCallsiteEntry  expected_named_arg;
+    MVMuint16         remaining_args;
     MVMuint16         remaining_positionals;
     MVMuint32         remaining_jumplabels;
     MVMuint32         reg_type_var;
@@ -110,6 +111,13 @@ static void ensure_no_remaining_positionals(Validator *val) {
 }
 
 
+static void ensure_no_remaining_args(Validator *val) {
+    if (val->remaining_args != 0)
+        fail(val, MSG(val, "callsite expects %" PRIu16 " more args"),
+                val->remaining_args);
+}
+
+
 MVM_STATIC_INLINE const MVMOpInfo * get_info(Validator *val, MVMuint16 opcode) {
     const MVMOpInfo *info;
 
@@ -291,8 +299,9 @@ next_operand:
 
 
 static void validate_lex_operand(Validator *val, MVMuint32 flags) {
+    MVMuint32 operand_type = flags & MVM_operand_type_mask;
     MVMuint16 lex_index, frame_index, i;
-    MVMuint32 lex_count;
+    MVMuint32 lex_count, lex_type;
     MVMStaticFrame *frame = val->frame;
 
     /* Two steps forward, two steps back to keep the error reporting happy,
@@ -312,11 +321,30 @@ static void validate_lex_operand(Validator *val, MVMuint32 flags) {
                     " more enclosing scopes"), i);
     }
 
+    if (!frame->body.fully_deserialized)
+        MVM_bytecode_finish_frame(val->tc, frame->body.cu, frame, 0);
+
     lex_count = frame->body.num_lexicals;
     if (lex_index >= lex_count)
         fail(val, MSG(val, "lexical operand index %" PRIu16
                 " out of range 0.. %" PRIu32), lex_index, lex_count - 1);
 
+    lex_type = frame->body.lexical_types[lex_index] << 3;
+
+    if (operand_type == MVM_operand_type_var) {
+        if (!val->reg_type_var) {
+            val->reg_type_var = lex_type;
+            goto next_operand;
+        }
+
+        operand_type = val->reg_type_var;
+    }
+
+    if (lex_type != operand_type)
+        fail(val, MSG(val, "operand type %i does not match lexical type %i"),
+                operand_type, lex_type);
+
+  next_operand:
     val->cur_op += 4;
 }
 
@@ -437,6 +465,8 @@ terminate_seq:
 static void validate_arg(Validator *val) {
     MVMCallsiteEntry flags;
 
+    val->remaining_args--;
+
     if (val->expected_named_arg) {
         flags = val->expected_named_arg;
         val->expected_named_arg = 0;
@@ -454,10 +484,13 @@ static void validate_arg(Validator *val) {
         switch (flags & ~MVM_CALLSITE_ARG_MASK) {
             case 0: /* positionals */
             case MVM_CALLSITE_ARG_FLAT:
-            case MVM_CALLSITE_ARG_FLAT_NAMED:
                 val->remaining_positionals--;
                 break;
 
+            case MVM_CALLSITE_ARG_FLAT_NAMED:
+                /* Nothing to do for this case. */
+                break;
+
             case MVM_CALLSITE_ARG_NAMED:
                 val->expected_named_arg = flags & MVM_CALLSITE_ARG_MASK;
                 goto named_arg;
@@ -527,7 +560,8 @@ static void validate_block(Validator *val) {
             index = GET_UI16(val->cur_op, -2);
             val->cur_call  = val->cu->body.callsites[index];
             val->cur_arg   = 0;
-            val->expected_named_arg    = 0;
+            val->expected_named_arg = 0;
+            val->remaining_args = val->cur_call->arg_count;
             val->remaining_positionals = val->cur_call->num_pos;
 
             break;
