@@ -69,10 +69,10 @@ MVMObject * MVM_proc_getenvhash(MVMThreadContext *tc) {
 
     while ((env = environ[pos++]) != NULL) {
 #ifndef _WIN32
-        MVMString    *str = MVM_string_utf8_decode(tc, instance->VMString, env, strlen(env));
+        MVMString    *str = MVM_string_utf8_c8_decode(tc, instance->VMString, env, strlen(env));
 #else
         char * const _env = ANSIToUTF8(acp, env);
-        MVMString    *str = MVM_string_utf8_decode(tc, instance->VMString, _env, strlen(_env));
+        MVMString    *str = MVM_string_utf8_c8_decode(tc, instance->VMString, _env, strlen(_env));
 #endif
 
         MVMuint32 index = MVM_string_index(tc, str, needle, 0);
@@ -112,7 +112,7 @@ MVMObject * MVM_proc_getenvhash(MVMThreadContext *tc) {
                 env_str = MVM_string_concatenate(tc, MVM_iterkey_s(tc, iter), equal); \
                 iterval = MVM_iterval(tc, iter); \
                 env_str = MVM_string_concatenate(tc, env_str, MVM_repr_get_str(tc, iterval)); \
-                _env[i++] = MVM_string_utf8_encode_C_string(tc, env_str); \
+                _env[i++] = MVM_string_utf8_c8_encode_C_string(tc, env_str); \
             } \
             _env[size] = NULL; \
         }); \
@@ -173,8 +173,8 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
     uv_stdio_container_t process_stdio[3];
     int i;
 
-    char * const cmdin = MVM_string_utf8_encode_C_string(tc, cmd);
-    char * const _cwd = MVM_string_utf8_encode_C_string(tc, cwd);
+    char * const cmdin = MVM_string_utf8_c8_encode_C_string(tc, cmd);
+    char * const _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
     const MVMuint64 size = MVM_repr_elems(tc, env);
     MVMIter * const iter = (MVMIter *)MVM_iter(tc, env);
     char **_env = MVM_malloc((size + 1) * sizeof(char *));
@@ -244,7 +244,7 @@ MVMint64 MVM_proc_spawn(MVMThreadContext *tc, MVMObject *argv, MVMString *cwd, M
     uv_stdio_container_t process_stdio[3];
     int i;
 
-    char   * const      _cwd = MVM_string_utf8_encode_C_string(tc, cwd);
+    char   * const      _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
     const MVMuint64     size = MVM_repr_elems(tc, env);
     MVMIter * const     iter = (MVMIter *)MVM_iter(tc, env);
     char              **_env = MVM_malloc((size + 1) * sizeof(char *));
@@ -255,7 +255,7 @@ MVMint64 MVM_proc_spawn(MVMThreadContext *tc, MVMObject *argv, MVMString *cwd, M
     i = 0;
     while(i < arg_size) {
         REPR(argv)->pos_funcs.at_pos(tc, STABLE(argv), argv, OBJECT_BODY(argv), i, &reg, MVM_reg_obj);
-        args[i++] = MVM_string_utf8_encode_C_string(tc, MVM_repr_get_str(tc, reg.o));
+        args[i++] = MVM_string_utf8_c8_encode_C_string(tc, MVM_repr_get_str(tc, reg.o));
     }
     args[arg_size] = NULL;
 
@@ -314,6 +314,12 @@ typedef struct {
     MVMint64 signal;
 } MVMIOAsyncProcessData;
 
+typedef enum {
+    STATE_UNSTARTED,
+    STATE_STARTED,
+    STATE_DONE
+} ProcessState;
+
 /* Info we convey about an async spawn task. */
 typedef struct {
     MVMThreadContext  *tc;
@@ -329,6 +335,7 @@ typedef struct {
     MVMuint32          seq_stdout;
     MVMuint32          seq_stderr;
     uv_stream_t       *stdin_handle;
+    ProcessState      state;
 } SpawnInfo;
 
 /* Info we convey about a write task. */
@@ -396,7 +403,7 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     /* Encode the string, or extract buf data. */
     if (wi->str_data) {
         MVMuint64 output_size_64;
-        output = MVM_string_utf8_encode(tc, wi->str_data, &output_size_64);
+        output = MVM_string_utf8_encode(tc, wi->str_data, &output_size_64, 1);
         output_size = (int)output_size_64;
     }
     else {
@@ -561,10 +568,30 @@ static const MVMAsyncTaskOps close_op_table = {
     NULL
 };
 
+static void deferred_close_perform(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data);
+
+static const MVMAsyncTaskOps deferred_close_op_table = {
+    deferred_close_perform,
+    NULL,
+    NULL,
+    NULL
+};
+
 static MVMint64 close_stdin(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOAsyncProcessData *handle_data = (MVMIOAsyncProcessData *)h->body.data;
     MVMAsyncTask          *spawn_task  = (MVMAsyncTask *)handle_data->async_task;
     SpawnInfo             *si          = spawn_task ? (SpawnInfo *)spawn_task->body.data : NULL;
+    if (si && si->state == STATE_UNSTARTED) {
+        MVMAsyncTask *task;
+        MVMROOT(tc, h, {
+            task = (MVMAsyncTask *)MVM_repr_alloc_init(tc,
+                tc->instance->boot_types.BOOTAsync);
+        });
+        task->body.ops  = &deferred_close_op_table;
+        task->body.data = si;
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+        return 0;
+    }
     if (si && si->stdin_handle) {
         MVMAsyncTask *task;
         MVMROOT(tc, h, {
@@ -579,6 +606,26 @@ static MVMint64 close_stdin(MVMThreadContext *tc, MVMOSHandle *h) {
     return 0;
 }
 
+static void deferred_close_perform(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
+    SpawnInfo *si = (SpawnInfo *) data;
+    MVMOSHandle *h = (MVMOSHandle *) si->handle;
+
+    if (si->state == STATE_UNSTARTED) {
+        MVMAsyncTask *task;
+        MVMROOT(tc, h, {
+            task = (MVMAsyncTask *)MVM_repr_alloc_init(tc,
+                tc->instance->boot_types.BOOTAsync);
+        });
+        task->body.ops  = &deferred_close_op_table;
+        task->body.data = si;
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+        return;
+    }
+    if (si->stdin_handle) {
+        close_stdin(tc, h);
+    }
+}
+
 /* IO ops table, for async process, populated with functions. */
 static const MVMIOAsyncWritable proc_async_writable = { write_str, write_bytes };
 static const MVMIOClosable      closable            = { close_stdin };
@@ -593,6 +640,8 @@ static const MVMIOOps proc_op_table = {
     NULL,
     NULL,
     NULL,
+    NULL,
+    NULL,
     proc_async_gc_mark,
     NULL
 };
@@ -634,6 +683,7 @@ static void async_spawn_on_exit(uv_process_t *req, MVMint64 exit_status, int ter
     /* when invoked via MVMIOOps, close_stdin is already wrapped in a mutex */
     os_handle = (MVMOSHandle *) si->handle;
     uv_mutex_lock(os_handle->body.mutex);
+    si->state = STATE_DONE;
     close_stdin(tc, os_handle);
     uv_mutex_unlock(os_handle->body.mutex);
 
@@ -693,9 +743,9 @@ static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf,
     else if (nread == UV_EOF) {
         MVMROOT(tc, t, {
         MVMROOT(tc, arr, {
-            MVMObject *minus_one = MVM_repr_box_int(tc,
-                tc->instance->boot_types.BOOTInt, -1);
-            MVM_repr_push_o(tc, arr, minus_one);
+            MVMObject *final = MVM_repr_box_int(tc,
+                tc->instance->boot_types.BOOTInt, seq_number);
+            MVM_repr_push_o(tc, arr, final);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
         });
@@ -784,7 +834,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         pipe->data = si;
         process_stdio[1].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
         process_stdio[1].data.stream = (uv_stream_t *)pipe;
-        si->ds_stdout                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0);
+        si->ds_stdout                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 1);
         stdout_pipe                  = pipe;
         stdout_cb                    = async_spawn_stdout_chars_read;
     }
@@ -807,7 +857,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         pipe->data = si;
         process_stdio[2].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
         process_stdio[2].data.stream = (uv_stream_t *)pipe;
-        si->ds_stderr                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0);
+        si->ds_stderr                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 1);
         stderr_pipe                  = pipe;
         stderr_cb                    = async_spawn_stderr_chars_read;
     }
@@ -841,6 +891,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     if (spawn_result) {
         MVMObject *error_cb = MVM_repr_at_key_o(tc, si->callbacks,
             tc->instance->str_consts.error);
+        si->state = STATE_DONE;
         if (!MVM_is_null(tc, error_cb)) {
             MVMROOT(tc, error_cb, {
             MVMROOT(tc, async_task, {
@@ -866,6 +917,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
 
         ready_cb = MVM_repr_at_key_o(tc, si->callbacks,
             tc->instance->str_consts.ready);
+        si->state = STATE_STARTED;
 
         if (!MVM_is_null(tc, ready_cb)) {
             MVMROOT(tc, ready_cb, {
@@ -893,7 +945,8 @@ static void spawn_cancel(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async
     MVMIOAsyncProcessData *apd     = (MVMIOAsyncProcessData *)handle->body.data;
     uv_process_t          *phandle = apd->handle;
 
-    /* If it didn't already end, try to kill it. */
+    /* If it didn't already end, try to kill it. exit_cb will clean up phandle
+     * should the signal lead to process exit. */
     if (phandle) {
 #ifdef _WIN32
         /* On Windows, make sure we use a signal that will actually work. */
@@ -901,8 +954,6 @@ static void spawn_cancel(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async
             apd->signal = SIGKILL;
 #endif
         uv_process_kill(phandle, (int)apd->signal);
-        uv_close((uv_handle_t *)phandle, spawn_async_close);
-        apd->handle = NULL;
     }
 }
 
@@ -977,13 +1028,13 @@ MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObje
     args = MVM_malloc((arg_size + 1) * sizeof(char *));
     for (i = 0; i < arg_size; i++) {
         REPR(argv)->pos_funcs.at_pos(tc, STABLE(argv), argv, OBJECT_BODY(argv), i, &reg, MVM_reg_obj);
-        args[i] = MVM_string_utf8_encode_C_string(tc, MVM_repr_get_str(tc, reg.o));
+        args[i] = MVM_string_utf8_c8_encode_C_string(tc, MVM_repr_get_str(tc, reg.o));
     }
     args[arg_size] = NULL;
     prog = args[0];
 
     /* Encode CWD. */
-    _cwd = MVM_string_utf8_encode_C_string(tc, cwd);
+    _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
 
     MVMROOT(tc, queue, {
     MVMROOT(tc, env, {
@@ -1013,6 +1064,7 @@ MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObje
         si->cwd         = _cwd;
         si->env         = _env;
         si->args        = args;
+        si->state       = STATE_UNSTARTED;
         MVM_ASSIGN_REF(tc, &(task->common.header), si->handle, handle);
         MVM_ASSIGN_REF(tc, &(task->common.header), si->callbacks, callbacks);
         task->body.data = si;
@@ -1090,7 +1142,7 @@ MVMnum64 MVM_proc_time_n(MVMThreadContext *tc) {
 MVMString * MVM_executable_name(MVMThreadContext *tc) {
     MVMInstance * const instance = tc->instance;
     if (instance->exec_name)
-        return MVM_string_utf8_decode(tc,
+        return MVM_string_utf8_c8_decode(tc,
             instance->VMString,
             instance->exec_name, strlen(instance->exec_name));
     else
@@ -1107,7 +1159,7 @@ MVMObject * MVM_proc_clargs(MVMThreadContext *tc) {
             const MVMint64 num_clargs = instance->num_clargs;
             MVMint64 count;
 
-            MVMString *prog_string = MVM_string_utf8_decode(tc,
+            MVMString *prog_string = MVM_string_utf8_c8_decode(tc,
                 instance->VMString,
                 instance->prog_name, strlen(instance->prog_name));
             MVMObject *boxed_str = MVM_repr_box_str(tc,
@@ -1116,7 +1168,7 @@ MVMObject * MVM_proc_clargs(MVMThreadContext *tc) {
 
             for (count = 0; count < num_clargs; count++) {
                 char *raw_clarg = instance->raw_clargs[count];
-                MVMString *string = MVM_string_utf8_decode(tc,
+                MVMString *string = MVM_string_utf8_c8_decode(tc,
                     instance->VMString, raw_clarg, strlen(raw_clarg));
                 boxed_str = MVM_repr_box_str(tc,
                     instance->boot_types.BOOTStr, string);
@@ -1129,7 +1181,7 @@ MVMObject * MVM_proc_clargs(MVMThreadContext *tc) {
             const MVMint64 num_clargs = instance->num_clargs;
             MVMint64 count;
 
-            MVMString *prog_string = MVM_string_utf8_decode(tc,
+            MVMString *prog_string = MVM_string_utf8_c8_decode(tc,
                 instance->VMString,
                 instance->prog_name, strlen(instance->prog_name));
             MVMObject *boxed_str = MVM_repr_box_str(tc,
@@ -1139,7 +1191,7 @@ MVMObject * MVM_proc_clargs(MVMThreadContext *tc) {
             for (count = 0; count < num_clargs; count++) {
                 char *raw_clarg = instance->raw_clargs[count];
                 char * const _tmp = ANSIToUTF8(acp, raw_clarg);
-                MVMString *string = MVM_string_utf8_decode(tc,
+                MVMString *string = MVM_string_utf8_c8_decode(tc,
                     instance->VMString, _tmp, strlen(_tmp));
                 MVM_free(_tmp);
                 boxed_str = MVM_repr_box_str(tc,
