@@ -1,6 +1,16 @@
 #include "moar.h"
 #include <platform/threads.h>
 
+#if defined(_MSC_VER)
+#define snprintf _snprintf
+#endif
+
+#ifndef _WIN32
+#  include <unistd.h>
+#else
+#  include <process.h>
+#endif
+
 #define init_mutex(loc, name) do { \
     if ((init_stat = uv_mutex_init(&loc)) < 0) { \
         fprintf(stderr, "MoarVM: Initialization of " name " mutex failed\n    %s\n", \
@@ -11,6 +21,48 @@
 
 static void setup_std_handles(MVMThreadContext *tc);
 
+static FILE *fopen_perhaps_with_pid(char *path, const char *mode) {
+    if (strstr(path, "%d")) {
+        MVMuint64 path_length = strlen(path);
+        MVMuint64 found_percents = 0;
+        MVMuint64 i;
+
+        /* Let's sanitize the format string a bit. Must only have
+         * a single printf-recognized directive. */
+        for (i = 0; i < path_length; i++) {
+            if (path[i] == '%') {
+                /* %% is all right. */
+                if (i + 1 < path_length && path[i + 1] == '%') {
+                    i++; continue;
+                }
+                found_percents++;
+            }
+        }
+        /* We expect to pass only a single argument to snprintf here;
+         * just bail out if there's more than one directive. */
+        if (found_percents > 1) {
+            return fopen(path, mode);
+        } else {
+            char *fixed_path = malloc(path_length + 16);
+            FILE *result;
+            MVMint64 pid;
+#ifdef _WIN32
+            pid = _getpid();
+#else
+            pid = getpid();
+#endif
+            /* We make the brave assumption that
+             * pids only go up to 16 characters. */
+            snprintf(fixed_path, path_length + 16, path, pid);
+            result = fopen(fixed_path, mode);
+            free(fixed_path);
+            return result;
+        }
+    } else {
+        return fopen(path, mode);
+    }
+}
+
 /* Create a new instance of the VM. */
 MVMInstance * MVM_vm_create_instance(void) {
     MVMInstance *instance;
@@ -106,6 +158,9 @@ MVMInstance * MVM_vm_create_instance(void) {
     /* Set up hll symbol tables mutex. */
     init_mutex(instance->mutex_hll_syms, "hll syms");
 
+    /* Initialize Unicode database */
+    MVM_unicode_init(instance->main_thread);
+
     /* Initialize string cclass handling. */
     MVM_string_cclass_init(instance->main_thread);
 
@@ -132,7 +187,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     init_mutex(instance->mutex_spesh_install, "spesh installations");
     spesh_log = getenv("MVM_SPESH_LOG");
     if (spesh_log && strlen(spesh_log))
-        instance->spesh_log_fh = fopen(spesh_log, "w");
+        instance->spesh_log_fh = fopen_perhaps_with_pid(spesh_log, "w");
     spesh_disable = getenv("MVM_SPESH_DISABLE");
     if (!spesh_disable || strlen(spesh_disable) == 0) {
         instance->spesh_enabled = 1;
@@ -157,15 +212,25 @@ MVMInstance * MVM_vm_create_instance(void) {
         instance->jit_enabled = 1;
     jit_log = getenv("MVM_JIT_LOG");
     if (jit_log && strlen(jit_log))
-        instance->jit_log_fh = fopen(jit_log, "w");
+        instance->jit_log_fh = fopen_perhaps_with_pid(jit_log, "w");
     jit_bytecode_dir = getenv("MVM_JIT_BYTECODE_DIR");
-    if (jit_bytecode_dir && strlen(jit_bytecode_dir))
+    if (jit_bytecode_dir && strlen(jit_bytecode_dir)) {
+        char *bytecode_map_name = MVM_malloc(strlen(jit_bytecode_dir) + strlen("/jit-map.txt") + 1);
+        sprintf(bytecode_map_name, "%s/jit-map.txt", jit_bytecode_dir);
+        instance->jit_bytecode_map = fopen(bytecode_map_name, "w");
         instance->jit_bytecode_dir = jit_bytecode_dir;
+        MVM_free(bytecode_map_name);
+    }
+    instance->jit_seq_nr = 0;
 
     /* Various kinds of debugging that can be enabled. */
     dynvar_log = getenv("MVM_DYNVAR_LOG");
-    if (dynvar_log && strlen(dynvar_log))
-        instance->dynvar_log_fh = fopen(dynvar_log, "w");
+    if (dynvar_log && strlen(dynvar_log)) {
+        instance->dynvar_log_fh = fopen_perhaps_with_pid(dynvar_log, "w");
+        fprintf(instance->dynvar_log_fh, "+ x 0 0 0 0 0 %llu\n", uv_hrtime());
+        fflush(instance->dynvar_log_fh);
+        instance->dynvar_log_lasttime = uv_hrtime();
+    }
     else
         instance->dynvar_log_fh = NULL;
     instance->nfa_debug_enabled = getenv("MVM_NFA_DEB") ? 1 : 0;
@@ -181,13 +246,13 @@ MVMInstance * MVM_vm_create_instance(void) {
         instance->cross_thread_write_logging = 0;
     }
 
-    /* Create std[in/out/err]. */
-    setup_std_handles(instance->main_thread);
-
     /* Set up NFG state mutation mutex. */
     instance->nfg = calloc(1, sizeof(MVMNFGState));
     init_mutex(instance->nfg->update_mutex, "NFG update mutex");
 
+    /* Create std[in/out/err]. */
+    setup_std_handles(instance->main_thread);
+
     /* Back to nursery allocation, now we're set up. */
     MVM_gc_allocate_gen2_default_clear(instance->main_thread);
 
@@ -224,7 +289,7 @@ void MVM_vm_run_file(MVMInstance *instance, const char *filename) {
     MVMROOT(tc, cu, {
         /* The call to MVM_string_utf8_decode() may allocate, invalidating the
            location cu->body.filename */
-        MVMString *const str = MVM_string_utf8_decode(tc, instance->VMString, filename, strlen(filename));
+        MVMString *const str = MVM_string_utf8_c8_decode(tc, instance->VMString, filename, strlen(filename));
         cu->body.filename = str;
 
         /* Run deserialization frame, if there is one. */
@@ -263,11 +328,43 @@ void MVM_vm_exit(MVMInstance *instance) {
         fclose(instance->spesh_log_fh);
     if (instance->jit_log_fh)
         fclose(instance->jit_log_fh);
+    if (instance->jit_bytecode_map)
+        fclose(instance->jit_bytecode_map);
+    if (instance->dynvar_log_fh) {
+        fprintf(instance->dynvar_log_fh, "- x 0 0 0 0 %lld %llu %llu\n", instance->dynvar_log_lasttime, uv_hrtime(), uv_hrtime());
+        fclose(instance->dynvar_log_fh);
+    }
 
     /* And, we're done. */
     exit(0);
 }
 
+static void cleanup_callsite_interns(MVMInstance *instance) {
+    int i;
+
+    for (i = 0; i < MVM_INTERN_ARITY_LIMIT; i++) {
+        int callsite_count = instance->callsite_interns->num_by_arity[i];
+        int j;
+
+        if (callsite_count) {
+            MVMCallsite **callsites = instance->callsite_interns->by_arity[i];
+
+            for (j = 0; j < callsite_count; j++) {
+                MVMCallsite *callsite = callsites[j];
+
+                if (MVM_callsite_is_common(callsite)) {
+                    continue;
+                }
+
+                MVM_callsite_destroy(callsite);
+            }
+
+            MVM_free(callsites);
+        }
+    }
+    MVM_free(instance->callsite_interns);
+}
+
 /* Destroys a VM instance. This must be called only from the main thread. It
  * should clear up all resources and free all memory; in practice, it falls
  * short of this goal at the moment. */
@@ -326,6 +423,16 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     /* Clean up multi cache addition mutex. */
     uv_mutex_destroy(&instance->mutex_multi_cache_add);
 
+    /* Clean up interned callsites */
+    uv_mutex_destroy(&instance->mutex_callsite_interns);
+    cleanup_callsite_interns(instance);
+
+    /* Clean up fixed size allocator */
+    MVM_fixed_size_destroy(instance->fsa);
+
+    /* Release this interpreter's hold on Unicode database */
+    MVM_unicode_release(instance->main_thread);
+
     /* Clean up spesh install mutex and close any log. */
     uv_mutex_destroy(&instance->mutex_spesh_install);
     if (instance->spesh_log_fh)
