@@ -27,10 +27,12 @@ static void uninline(MVMThreadContext *tc, MVMFrame *f, MVMSpeshCandidate *cand,
                 MVM_string_utf8_encode_C_string(tc, usf->body.cuuid));*/
 
             /* Copy the locals and lexicals into place. */
-            memcpy(uf->work, f->work + cand->inlines[i].locals_start,
-                usf->body.num_locals * sizeof(MVMRegister));
-            memcpy(uf->env, f->env + cand->inlines[i].lexicals_start,
-                usf->body.num_lexicals * sizeof(MVMRegister));
+            if (usf->body.num_locals)
+                memcpy(uf->work, f->work + cand->inlines[i].locals_start,
+                    usf->body.num_locals * sizeof(MVMRegister));
+            if (usf->body.num_lexicals)
+                memcpy(uf->env, f->env + cand->inlines[i].lexicals_start,
+                    usf->body.num_lexicals * sizeof(MVMRegister));
 
             /* Did we already uninline a frame? */
             if (last_uninlined) {
@@ -53,7 +55,7 @@ static void uninline(MVMThreadContext *tc, MVMFrame *f, MVMSpeshCandidate *cand,
                     uf->return_value = uf->work + last_res_reg;
 
                 /* Set up last uninlined's caller to us. */
-                last_uninlined->caller = MVM_frame_inc_ref(tc, uf);
+                last_uninlined->caller = MVM_frame_inc_ref_by_frame(tc, uf);
             }
             else {
                 /* First uninlined frame. Are we in the middle of the call
@@ -62,7 +64,7 @@ static void uninline(MVMThreadContext *tc, MVMFrame *f, MVMSpeshCandidate *cand,
                     /* Tweak the callee's caller to the uninlined frame, not
                      * the frame holding the inlinings. */
                     MVMFrame *orig_caller = callee->caller;
-                    callee->caller = MVM_frame_inc_ref(tc, uf);
+                    callee->caller = MVM_frame_inc_ref_by_frame(tc, uf);
                     MVM_frame_dec_ref(tc, orig_caller);
 
                     /* Copy over the return location. */
@@ -113,7 +115,7 @@ static void uninline(MVMThreadContext *tc, MVMFrame *f, MVMSpeshCandidate *cand,
             f->return_value = f->work + last_res_reg;
 
         /* Set up inliner as the caller, given we now have a direct inline. */
-        last_uninlined->caller = MVM_frame_inc_ref(tc, f);
+        last_uninlined->caller = MVM_frame_inc_ref_by_frame(tc, f);
     }
     else {
         /* Weren't in an inline after all. What kind of deopt? */
@@ -136,7 +138,7 @@ static MVMint32 find_deopt_target(MVMThreadContext *tc, MVMFrame *f, MVMint32 de
             return f->spesh_cand->deopts[i];
         }
     }
-    MVM_exception_throw_adhoc(tc, "find_deopt_target failed for %s (%s)",
+    MVM_oops(tc, "find_deopt_target failed for %s (%s)",
         MVM_string_utf8_encode_C_string(tc, tc->cur_frame->static_info->body.name),
         MVM_string_utf8_encode_C_string(tc, tc->cur_frame->static_info->body.cuuid));
 }
@@ -189,7 +191,7 @@ void MVM_spesh_deopt_one(MVMThreadContext *tc) {
         deopt_frame(tc, tc->cur_frame, deopt_offset, deopt_target);
     }
     else {
-        MVM_exception_throw_adhoc(tc, "deopt_one failed for %s (%s)",
+        MVM_oops(tc, "deopt_one failed for %s (%s)",
             MVM_string_utf8_encode_C_string(tc, tc->cur_frame->static_info->body.name),
             MVM_string_utf8_encode_C_string(tc, tc->cur_frame->static_info->body.cuuid));
     }
@@ -204,7 +206,7 @@ void MVM_spesh_deopt_one_direct(MVMThreadContext *tc, MVMint32 deopt_offset,
     if (f->effective_bytecode != f->static_info->body.bytecode) {
         deopt_frame(tc, tc->cur_frame, deopt_offset, deopt_target);
     } else {
-        MVM_exception_throw_adhoc(tc, "deopt_one_direct failed for %s (%s)",
+        MVM_oops(tc, "deopt_one_direct failed for %s (%s)",
             MVM_string_utf8_encode_C_string(tc, tc->cur_frame->static_info->body.name),
             MVM_string_utf8_encode_C_string(tc, tc->cur_frame->static_info->body.cuuid));
     }
