@@ -241,19 +241,19 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                             break;
                         case MVM_operand_coderef: {
                             MVMCodeBody *body = &((MVMCode*)g->sf->body.cu->body.coderefs[cur_ins->operands[i].coderef_idx])->body;
-                            MVMBytecodeAnnotation *ann = MVM_bytecode_resolve_annotation(tc, &body->sf->body, 0);
+                            MVMBytecodeAnnotation *anno = MVM_bytecode_resolve_annotation(tc, &body->sf->body, 0);
 
                             append(ds, "coderef(");
 
-                            if (ann) {
-                                char *filestr = MVM_string_utf8_encode_C_string(tc, g->sf->body.cu->body.strings[ann->filename_string_heap_index]);
-                                appendf(ds, "%s:%d%s)", filestr, ann->line_number, body->outer ? " (closure)" : "");
+                            if (anno) {
+                                char *filestr = MVM_string_utf8_encode_C_string(tc, g->sf->body.cu->body.strings[anno->filename_string_heap_index]);
+                                appendf(ds, "%s:%d%s)", filestr, anno->line_number, body->outer ? " (closure)" : "");
                                 MVM_free(filestr);
                             } else {
                                 append(ds, "??\?)");
                             }
 
-                            MVM_free(ann);
+                            MVM_free(anno);
                             break;
                         }
                         default:
@@ -334,6 +334,9 @@ static void dump_facts(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
             if (flags & 4096) {
                 append(ds, " MgWLG");
             }
+            if (flags & 8192) {
+                append(ds, " KRWCn");
+            }
             if (g->facts[i][j].writer && g->facts[i][j].writer->info->opcode == MVM_SSA_PHI) {
                 appendf(ds, " (merged from %d regs)", g->facts[i][j].writer->info->num_operands - 1);
             }
@@ -461,6 +464,12 @@ static void dump_arg_guards(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g)
         case MVM_SPESH_GUARD_DC_TYPE:
             appendf(ds, "  deconted_type(%d, %p)\n", guard->slot, guard->match);
             break;
+        case MVM_SPESH_GUARD_DC_CONC_RW:
+            appendf(ds, "  deconted_concrete_rw(%d)\n", guard->slot);
+            break;
+        case MVM_SPESH_GUARD_DC_TYPE_RW:
+            appendf(ds, "  deconted_type_rw(%d, %p)\n", guard->slot, guard->match);
+            break;
         }
     }
     append(ds, "\n");
