@@ -22,20 +22,13 @@ static void check_strand_sanity(MVMThreadContext *tc, MVMString *s) {
 #define STRAND_CHECK(tc, s)
 #endif
 
-/* Checks a string is not null or non-concrete and throws if so. */
-MVM_STATIC_INLINE void MVM_string_check_arg(MVMThreadContext *tc, MVMString *s, const char *operation) {
-    if (!s || !IS_CONCRETE(s))
-        MVM_exception_throw_adhoc(tc, "%s requires a concrete string, but got %s",
-            operation, s ? "a type object" : "null");
-}
-
 /* Allocates strand storage. */
 static MVMStringStrand * allocate_strands(MVMThreadContext *tc, MVMuint16 num_strands) {
     return MVM_malloc(num_strands * sizeof(MVMStringStrand));
 }
 
 /* Copies strands from one strand string to another. */
-static void copy_strands(MVMThreadContext *tc, MVMString *from, MVMuint16 from_offset,
+static void copy_strands(MVMThreadContext *tc, const MVMString *from, MVMuint16 from_offset,
         MVMString *to, MVMuint16 to_offset, MVMuint16 num_strands) {
     assert(from->body.storage_type == MVM_STRING_STRAND);
     assert(to->body.storage_type == MVM_STRING_STRAND);
@@ -85,7 +78,7 @@ static MVMString * re_nfg(MVMThreadContext *tc, MVMString *in) {
 
     /* Iterate codepoints and normalizer. */
     MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
-    MVM_string_ci_init(tc, &ci, in);
+    MVM_string_ci_init(tc, &ci, in, 0);
     while (MVM_string_ci_has_more(tc, &ci)) {
         MVMGrapheme32 g;
         ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, MVM_string_ci_get_codepoint(tc, &ci), &g);
@@ -546,15 +539,47 @@ MVMint64 MVM_string_equal_at_ignore_case(MVMThreadContext *tc, MVMString *a, MVM
     return MVM_string_equal_at(tc, lca, lcb, offset);
 }
 
+MVMGrapheme32 MVM_string_ord_at(MVMThreadContext *tc, MVMString *s, MVMint64 offset) {
+    MVMStringIndex agraphs;
+    MVMGrapheme32 g;
+
+    MVM_string_check_arg(tc, s, "grapheme_at");
+
+    agraphs = MVM_string_graphs(tc, s);
+    if (offset < 0 || offset >= agraphs)
+	return -1;
+
+    g = MVM_string_get_grapheme_at_nocheck(tc, s, offset);
+
+    return g >= 0 ? g : MVM_nfg_get_synthetic_info(tc, g)->base;
+}
+
 MVMGrapheme32 MVM_string_ord_basechar_at(MVMThreadContext *tc, MVMString *s, MVMint64 offset) {
-    MVMGrapheme32 g = MVM_string_get_grapheme_at(tc, s, offset);
+    MVMStringIndex agraphs;
+    MVMGrapheme32 g;
     MVMNormalizer norm;
     MVMint32 ready;
-    MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
-    ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, g, &g);
-    MVM_unicode_normalizer_eof(tc, &norm);
-    if (!ready)
-        g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+
+    MVM_string_check_arg(tc, s, "grapheme_at");
+
+    agraphs = MVM_string_graphs(tc, s);
+    if (offset < 0 || offset >= agraphs)
+	return -1;  /* fixes RT #126771 */
+
+    g = MVM_string_get_grapheme_at_nocheck(tc, s, offset);
+
+    if (g < 0) {
+	MVMNFGSynthetic *si = MVM_nfg_get_synthetic_info(tc, g);
+	g = si->base;
+    }
+    else {
+	MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
+	ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, g, &g);
+	MVM_unicode_normalizer_eof(tc, &norm);
+	if (!ready)
+	    g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+    }
+
     return g;
 }
 
@@ -617,64 +642,160 @@ MVMint64 MVM_string_index_of_grapheme(MVMThreadContext *tc, MVMString *a, MVMGra
     return -1;
 }
 
-#define change_case_iterate(member, dest_member, dest_size) \
-for (i = string->body.member + start; i < string->body.member + start + length; ) { \
-    if (dest->body.graphs == state->size) { \
-        if (!state->size) state->size = 16; \
-        else state->size *= 2; \
-        dest->body.dest_member = MVM_realloc(dest->body.dest_member, \
-            state->size * sizeof(dest_size)); \
-    } \
-    dest->body.dest_member[dest->body.graphs++] = \
-        MVM_unicode_get_case_change(tc, (MVMCodepoint32) *i++, state->case_change_type); \
-}
-
 /* Case change functions. */
-#define case_change_func(funcname, type, error) \
-MVMString * funcname(MVMThreadContext *tc, MVMString *s) { \
-    MVMint64 sgraphs; \
-    MVM_string_check_arg(tc, s, error); \
-    sgraphs = MVM_string_graphs(tc, s); \
-    if (sgraphs) { \
-        MVMString *result; \
-        MVMGraphemeIter gi; \
-        MVMGrapheme32 *result_buf = MVM_malloc(sgraphs * sizeof(MVMGrapheme32)); \
-        MVMint32 changed = 0; \
-        MVMint64 i = 0; \
-        MVM_string_gi_init(tc, &gi, s); \
-        while (i < sgraphs) { \
-            MVMGrapheme32 before = MVM_string_gi_get_grapheme(tc, &gi); \
-            MVMGrapheme32 after  = before >= 0 \
-                ? MVM_unicode_get_case_change(tc, before, type) \
-                : MVM_nfg_get_case_change(tc, before, type); \
-            result_buf[i++]      = after; \
-            if (before != after) \
-                changed = 1; \
-        } \
-        if (changed) { \
-            result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString); \
-            result->body.num_graphs      = sgraphs; \
-            result->body.storage_type    = MVM_STRING_GRAPHEME_32; \
-            result->body.storage.blob_32 = result_buf; \
-            return result; \
-        } \
-        else { \
-            MVM_free(result_buf); \
-        } \
-    } \
-    STRAND_CHECK(tc, s); \
-    return s; \
+static MVMint64 grapheme_is_cclass(MVMThreadContext *tc, MVMint64 cclass, MVMGrapheme32 g);
+static MVMString * do_case_change(MVMThreadContext *tc, MVMString *s, MVMint32 type, char *error) {
+    MVMint64 sgraphs;
+    MVM_string_check_arg(tc, s, error);
+    sgraphs = MVM_string_graphs(tc, s);
+    if (sgraphs) {
+        MVMString *result;
+        MVMGraphemeIter gi;
+        MVMint64 result_graphs = sgraphs;
+        MVMGrapheme32 *result_buf = MVM_malloc(result_graphs * sizeof(MVMGrapheme32));
+        MVMint32 changed = 0;
+        MVMint64 i = 0;
+        MVM_string_gi_init(tc, &gi, s);
+        while (MVM_string_gi_has_more(tc, &gi)) {
+            MVMGrapheme32 g = MVM_string_gi_get_grapheme(tc, &gi);
+          peeked:
+            if (g == 0x03A3) {
+                /* Greek sigma needs special handling when lowercased. */
+                switch (type) {
+                    case MVM_unicode_case_change_type_upper:
+                    case MVM_unicode_case_change_type_title:
+                        result_buf[i++] = g;
+                        break;
+                    case MVM_unicode_case_change_type_lower:
+                        changed = 1;
+                        if (i == 0) {
+                            /* Start of string, so not final. */
+                            result_buf[i++] = 0x03C3;
+                        }
+                        else if (!grapheme_is_cclass(tc, MVM_CCLASS_ALPHABETIC, result_buf[i - 1])) {
+                            /* Previous char is not a letter; not final (as has
+                             * to be at end of a word and not only thing in a
+                             * word). */
+                            result_buf[i++] = 0x03C3;
+                        }
+                        else if (!MVM_string_gi_has_more(tc, &gi)) {
+                            /* End of string. We only reach here if we have a
+                             * letter before us, so it must be final. */
+                            result_buf[i++] = 0x03C2;
+                        }
+                        else {
+                            /* Letter before us, something ahead of us. Need to
+                             * peek ahead to see if it's a letter, to decide if
+                             * we have final sigma or not. */
+                            g = MVM_string_gi_get_grapheme(tc, &gi);
+                            if (grapheme_is_cclass(tc, MVM_CCLASS_ALPHABETIC, g))
+                                result_buf[i++] = 0x03C3;
+                            else
+                                result_buf[i++] = 0x03C2;
+                            goto peeked;
+                        }
+                        break;
+                    case MVM_unicode_case_change_type_fold:
+                        result_buf[i++] = 0x03C3;
+                        changed = 1;
+                        break;
+                }
+            }
+            else if (g >= 0) {
+                const MVMCodepoint *result_cps;
+                MVMuint32 num_result_cps = MVM_unicode_get_case_change(tc,
+                    g, type, &result_cps);
+                if (num_result_cps == 0) {
+                    result_buf[i++] = g;
+                }
+                else if (num_result_cps == 1) {
+                    result_buf[i++] = *result_cps;
+                    changed = 1;
+                }
+                else {
+                    /* To maintain NFG, we need to re-normalize when we get an
+                     * expansion. */
+                    MVMNormalizer norm;
+                    MVMint32 num_result_graphs;
+                    MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
+                    MVM_unicode_normalizer_push_codepoints(tc, &norm, result_cps, num_result_cps);
+                    MVM_unicode_normalizer_eof(tc, &norm);
+                    num_result_graphs = MVM_unicode_normalizer_available(tc, &norm);
+
+                    /* Make space for any extra graphemes. */
+                    if (num_result_graphs > 1) {
+                        result_graphs += num_result_graphs - 1;
+                        result_buf = MVM_realloc(result_buf,
+                            result_graphs * sizeof(MVMGrapheme32));
+                    }
+
+                    /* Copy resulting graphemes. */
+                    while (num_result_graphs > 0) {
+                        result_buf[i++] = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+                        num_result_graphs--;
+                    }
+                    changed = 1;
+
+                    /* Clean up normalizer (we could init one per transform
+                     * and keep it around in the future, if we find it's a
+                     * worthwhile gain). */
+                    MVM_unicode_normalizer_cleanup(tc, &norm);
+                }
+            }
+            else {
+                MVMGrapheme32 *transformed;
+                MVMint32 num_transformed = MVM_nfg_get_case_change(tc, g, type, &transformed);
+                if (num_transformed == 0) {
+                    result_buf[i++] = g;
+                }
+                else if (num_transformed == 1) {
+                    result_buf[i++] = *transformed;
+                    changed = 1;
+                }
+                else {
+                    MVMuint32 j;
+                    result_graphs += num_transformed - 1;
+                    result_buf = MVM_realloc(result_buf,
+                        result_graphs * sizeof(MVMGrapheme32));
+                    for (j = 0; j < num_transformed; j++)
+                        result_buf[i++] = transformed[j];
+                    changed = 1;
+                }
+            }
+        }
+        if (changed) {
+            result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
+            result->body.num_graphs      = result_graphs;
+            result->body.storage_type    = MVM_STRING_GRAPHEME_32;
+            result->body.storage.blob_32 = result_buf;
+            return result;
+        }
+        else {
+            MVM_free(result_buf);
+        }
+    }
+    STRAND_CHECK(tc, s);
+    return s;
+}
+MVMString * MVM_string_uc(MVMThreadContext *tc, MVMString *s) {
+    return do_case_change(tc, s, MVM_unicode_case_change_type_upper, "uc");
+}
+MVMString * MVM_string_lc(MVMThreadContext *tc, MVMString *s) {
+    return do_case_change(tc, s, MVM_unicode_case_change_type_lower, "lc");
+}
+MVMString * MVM_string_tc(MVMThreadContext *tc, MVMString *s) {
+    return do_case_change(tc, s, MVM_unicode_case_change_type_title, "tc");
+}
+MVMString * MVM_string_fc(MVMThreadContext *tc, MVMString *s) {
+    return do_case_change(tc, s, MVM_unicode_case_change_type_fold, "fc");
 }
-case_change_func(MVM_string_uc, MVM_unicode_case_change_type_upper, "uc")
-case_change_func(MVM_string_lc, MVM_unicode_case_change_type_lower, "lc")
-case_change_func(MVM_string_tc, MVM_unicode_case_change_type_title, "tc")
 
 /* Decodes a C buffer to an MVMString, dependent on the encoding type flag. */
 MVMString * MVM_string_decode(MVMThreadContext *tc,
-        MVMObject *type_object, char *Cbuf, MVMint64 byte_length, MVMint64 encoding_flag) {
+        const MVMObject *type_object, char *Cbuf, MVMint64 byte_length, MVMint64 encoding_flag) {
     switch(encoding_flag) {
         case MVM_encoding_type_utf8:
-            return MVM_string_utf8_decode(tc, type_object, Cbuf, byte_length);
+            return MVM_string_utf8_decode_strip_bom(tc, type_object, Cbuf, byte_length);
         case MVM_encoding_type_ascii:
             return MVM_string_ascii_decode(tc, type_object, Cbuf, byte_length);
         case MVM_encoding_type_latin1:
@@ -683,34 +804,38 @@ MVMString * MVM_string_decode(MVMThreadContext *tc,
             return MVM_string_utf16_decode(tc, type_object, Cbuf, byte_length);
         case MVM_encoding_type_windows1252:
             return MVM_string_windows1252_decode(tc, type_object, Cbuf, byte_length);
+        case MVM_encoding_type_utf8_c8:
+            return MVM_string_utf8_c8_decode(tc, type_object, Cbuf, byte_length);
         default:
             MVM_exception_throw_adhoc(tc, "invalid encoding type flag: %"PRId64, encoding_flag);
     }
-    return NULL;
 }
 
 /* Encodes an MVMString to a C buffer, dependent on the encoding type flag */
-char * MVM_string_encode(MVMThreadContext *tc, MVMString *s, MVMint64 start, MVMint64 length, MVMuint64 *output_size, MVMint64 encoding_flag) {
+char * MVM_string_encode(MVMThreadContext *tc, MVMString *s, MVMint64 start,
+        MVMint64 length, MVMuint64 *output_size, MVMint64 encoding_flag,
+        MVMString *replacement, MVMint32 translate_newlines) {
     switch(encoding_flag) {
         case MVM_encoding_type_utf8:
-            return MVM_string_utf8_encode_substr(tc, s, output_size, start, length);
+            return MVM_string_utf8_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_ascii:
-            return MVM_string_ascii_encode_substr(tc, s, output_size, start, length);
+            return MVM_string_ascii_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_latin1:
-            return MVM_string_latin1_encode_substr(tc, s, output_size, start, length);
+            return MVM_string_latin1_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_utf16:
-            return MVM_string_utf16_encode_substr(tc, s, output_size, start, length);
+            return MVM_string_utf16_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_windows1252:
-            return MVM_string_windows1252_encode_substr(tc, s, output_size, start, length);
+            return MVM_string_windows1252_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
+        case MVM_encoding_type_utf8_c8:
+            return MVM_string_utf8_c8_encode_substr(tc, s, output_size, start, length, replacement);
         default:
             MVM_exception_throw_adhoc(tc, "invalid encoding type flag: %"PRId64, encoding_flag);
     }
-    return NULL;
 }
 
 /* Encodes a string, and writes the encoding string into the supplied Buf
  * instance, which should be an integer array with MVMArray REPR. */
-void MVM_string_encode_to_buf(MVMThreadContext *tc, MVMString *s, MVMString *enc_name, MVMObject *buf) {
+void MVM_string_encode_to_buf(MVMThreadContext *tc, MVMString *s, MVMString *enc_name, MVMObject *buf, MVMString *replacement) {
     MVMuint64 output_size;
     MVMuint8 *encoded;
     MVMArrayREPRData *buf_rd;
@@ -744,7 +869,7 @@ void MVM_string_encode_to_buf(MVMThreadContext *tc, MVMString *s, MVMString *enc
     MVMROOT(tc, s, {
         const MVMuint8 encoding_flag = MVM_string_find_encoding(tc, enc_name);
         encoded = (MVMuint8 *)MVM_string_encode(tc, s, 0, MVM_string_graphs(tc, s), &output_size,
-            encoding_flag);
+            encoding_flag, replacement, 0);
     });
     });
 
@@ -938,28 +1063,38 @@ MVMString * MVM_string_join(MVMThreadContext *tc, MVMString *separator, MVMObjec
 
             /* Add separator if needed. */
             if (i > 0) {
-                if (!concats_stable)
-                    /* Already stable; no more checks. */;
-                else if (!MVM_nfg_is_concat_stable(tc, pieces[i - 1], separator))
-                    concats_stable = 0;
-                else if (!MVM_nfg_is_concat_stable(tc, separator, piece))
-                    concats_stable = 0;
+                if (sgraphs) {
+                    if (!concats_stable)
+                        /* Already stable; no more checks. */;
+                    else if (!MVM_nfg_is_concat_stable(tc, pieces[i - 1], separator))
+                        concats_stable = 0;
+                    else if (!MVM_nfg_is_concat_stable(tc, separator, piece))
+                        concats_stable = 0;
 
-                switch (separator->body.storage_type) {
-                case MVM_STRING_GRAPHEME_32:
-                    memcpy(
-                        result->body.storage.blob_32 + position,
-                        separator->body.storage.blob_32,
-                        sgraphs * sizeof(MVMGrapheme32));
-                    position += sgraphs;
-                    break;
-                /* XXX Can special-case 8-bit NFG and ASCII here too. */
-                default:
-                    MVM_string_gi_init(tc, &gi, separator);
-                    while (MVM_string_gi_has_more(tc, &gi))
-                        result->body.storage.blob_32[position++] =
-                            MVM_string_gi_get_grapheme(tc, &gi);
-                    break;
+                    switch (separator->body.storage_type) {
+                    case MVM_STRING_GRAPHEME_32:
+                        memcpy(
+                            result->body.storage.blob_32 + position,
+                            separator->body.storage.blob_32,
+                            sgraphs * sizeof(MVMGrapheme32));
+                        position += sgraphs;
+                        break;
+                    /* XXX Can special-case 8-bit NFG and ASCII here too. */
+                    default:
+                        MVM_string_gi_init(tc, &gi, separator);
+                        while (MVM_string_gi_has_more(tc, &gi))
+                            result->body.storage.blob_32[position++] =
+                                MVM_string_gi_get_grapheme(tc, &gi);
+                        break;
+                    }
+                }
+                else {
+                    /* Separator has no graphemes, so NFG stability check
+                     * should consider pieces. */
+                    if (!concats_stable)
+                        /* Already stable; no more checks. */;
+                    else if (!MVM_nfg_is_concat_stable(tc, pieces[i - 1], piece))
+                        concats_stable = 0;
                 }
             }
 
@@ -1102,6 +1237,7 @@ MVMString * MVM_string_escape(MVMThreadContext *tc, MVMString *s) {
     MVMStringIndex  bpos    = 0;
     MVMStringIndex  sgraphs, balloc;
     MVMGrapheme32  *buffer;
+    MVMGrapheme32 crlf;
 
     MVM_string_check_arg(tc, s, "escape");
 
@@ -1109,6 +1245,8 @@ MVMString * MVM_string_escape(MVMThreadContext *tc, MVMString *s) {
     balloc  = sgraphs;
     buffer  = MVM_malloc(sizeof(MVMGrapheme32) * balloc);
 
+    crlf = MVM_nfg_crlf_grapheme(tc);
+
     for (; spos < sgraphs; spos++) {
         MVMGrapheme32 graph = MVM_string_get_grapheme_at_nocheck(tc, s, spos);
         MVMGrapheme32 esc   = 0;
@@ -1131,6 +1269,16 @@ MVMString * MVM_string_escape(MVMThreadContext *tc, MVMString *s) {
             buffer[bpos++] = '\\';
             buffer[bpos++] = esc;
         }
+        else if (graph == crlf) {
+            if (bpos + 4 > balloc) {
+                balloc += 32;
+                buffer = MVM_realloc(buffer, sizeof(MVMGrapheme32) * balloc);
+            }
+            buffer[bpos++] = '\\';
+            buffer[bpos++] = 'r';
+            buffer[bpos++] = '\\';
+            buffer[bpos++] = 'n';
+        }
         else {
             if (bpos + 1 > balloc) {
                 balloc += 32;
@@ -1451,7 +1599,7 @@ static MVMint64 grapheme_is_cclass(MVMThreadContext *tc, MVMint64 cclass, MVMGra
         }
 
         case MVM_CCLASS_PRINTING: {
-            return !(cp >= 0 && cp < 32) || (cp >= 127 && cp < 160);
+            return !((cp >= 0 && cp < 32) || (cp >= 127 && cp < 160));
         }
 
         case MVM_CCLASS_PUNCTUATION:
@@ -1549,6 +1697,7 @@ static MVMString *encoding_ascii_name        = NULL;
 static MVMString *encoding_latin1_name       = NULL;
 static MVMString *encoding_utf16_name        = NULL;
 static MVMString *encoding_windows1252_name  = NULL;
+static MVMString *encoding_utf8_c8_name      = NULL;
 MVMuint8 MVM_string_find_encoding(MVMThreadContext *tc, MVMString *name) {
     MVM_string_check_arg(tc, name, "find encoding");
     if (!encoding_name_init) {
@@ -1563,6 +1712,8 @@ MVMuint8 MVM_string_find_encoding(MVMThreadContext *tc, MVMString *name) {
         MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_utf16_name);
         encoding_windows1252_name = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "windows-1252");
         MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_windows1252_name);
+        encoding_utf8_c8_name     = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "utf8-c8");
+        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_utf8_c8_name);
         encoding_name_init   = 1;
         MVM_gc_allocate_gen2_default_clear(tc);
     }
@@ -1582,6 +1733,9 @@ MVMuint8 MVM_string_find_encoding(MVMThreadContext *tc, MVMString *name) {
     else if (MVM_string_equal(tc, name, encoding_utf16_name)) {
         return MVM_encoding_type_utf16;
     }
+    else if (MVM_string_equal(tc, name, encoding_utf8_c8_name)) {
+        return MVM_encoding_type_utf8_c8;
+    }
     else {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
