@@ -62,6 +62,7 @@ sub main {
     enumerated_property('Blocks', 'Block', { No_Block => 0 }, 1, 1);
     enumerated_property('extracted/DerivedDecompositionType', 'Decomposition_Type', { None => 0 }, 1, 1);
     CaseFolding();
+    SpecialCasing();
     enumerated_property('DerivedAge',
         'Age', { Unassigned => 0 }, 1, 1);
     binary_props('DerivedCoreProperties');
@@ -88,12 +89,12 @@ sub main {
     NamedSequences();
     binary_props('PropList');
     enumerated_property('Scripts', 'Script', { Unknown => 0 }, 1, 1);
-    # XXX SpecialCasing.txt # haven't decided how to do it
     # XXX StandardizedVariants.txt # no clue what this is
     break_property('Grapheme', 'Grapheme_Cluster_Break');
     break_property('Sentence', 'Sentence_Break');
   skip_most:
     break_property('Word', 'Word_Break');
+    tweak_nfg_qc();
 
     # Allocate all the things
     progress "done.\nallocating bitfield...";
@@ -851,7 +852,7 @@ static int block_compare(const void *a, const void *b) {
 MVMint32 MVM_unicode_is_in_block(MVMThreadContext *tc, MVMString *str, MVMint64 pos, MVMString *block_name) {
     MVMGrapheme32 ord = MVM_string_get_grapheme_at_nocheck(tc, str, pos);
     MVMuint64 size;
-    char *bname = MVM_string_ascii_encode(tc, block_name, &size);
+    char *bname = MVM_string_ascii_encode(tc, block_name, &size, 0);
     MVMint32 in_block = 0;
 
     struct UnicodeBlock *block = bsearch(&ord, unicode_blocks, sizeof(unicode_blocks) / sizeof(struct UnicodeBlock), sizeof(struct UnicodeBlock), block_compare);
@@ -1368,6 +1369,7 @@ sub UnicodeData {
         my $point = {
             code_str => $code_str,
             name => $name,
+            gencat_name => $gencat,
             General_Category => $general_categories->{enum}->{$gencat},
             Canonical_Combining_Class => $ccclasses->{enum}->{$ccclass},
             Bidi_Class => $bidi_classes->{enum}->{$bidiclass},
@@ -1378,6 +1380,7 @@ sub UnicodeData {
             NFC_QC => 1, # which will be unset as appropriate
             NFKD_QC => 1,
             NFKC_QC => 1,
+            NFG_QC => 1,
             code => $code,
             Any => 1
         };
@@ -1478,6 +1481,34 @@ sub CaseFolding {
     $db_sections->{BBB_CaseFolding_grows} = $grows_out;
 }
 
+sub SpecialCasing {
+    my $count = 1;
+    my @entries;
+    each_line('SpecialCasing', sub { $_ = shift;
+        s/#.+//;
+        my ($code, $lower, $title, $upper, $cond) = split /\s*;\s*/;
+        return if $cond;
+        sub threesome {
+            my @things = split ' ', shift;
+            push @things, 0 while @things < 3;
+            join ", ", map { "0x$_" } @things
+        }
+        push @entries, "{ { " . threesome($upper) .
+                       " }, { " . threesome($lower) .
+                       " }, { " . threesome($title) .
+                       " } }";
+        $points_by_hex->{$code}->{Special_Casing} = $count;
+        $count++;
+    });
+    my $out = "static const MVMint32 SpecialCasing_table[$count][3][3] = {\n    {0x0,0x0,0x0},\n    "
+        .stack_lines(\@entries, ",", ",\n    ", 0, $wrap_to_columns)."\n};";
+    my $bit_width = least_int_ge_lg2($count);
+    my $index_base = { bit_width => $bit_width };
+    register_enumerated_property('Special_Casing', $index_base);
+    $estimated_total_bytes += $count * 4 * 3 * 3;
+    $db_sections->{BBB_SpecialCasing} = $out;
+}
+
 sub DerivedNormalizationProps {
     my $binary = {
         Full_Composition_Exclusion => 1,
@@ -1490,7 +1521,8 @@ sub DerivedNormalizationProps {
     register_binary_property($_) for ((keys %$binary),(keys %$inverted_binary));
     my $trinary = {
         NFC_QC => 1,
-        NFKC_QC => 1
+        NFKC_QC => 1,
+        NFG_QC => 1,
     };
     my $trinary_values = { 'N' => 0, 'Y' => 1, 'M' => 2 };
     register_enumerated_property($_, { enum => $trinary_values, bit_width => 2, 'keys' => ['N','Y','M'] }) for (keys %$trinary);
@@ -1518,6 +1550,15 @@ sub DerivedNormalizationProps {
             my $point = shift;
             $point->{$property_name} = $value;
         });
+
+        # If it's the NFC_QC property, then use this as the default value for
+        # NFG_QC also.
+        if ($property_name eq 'NFC_QC') {
+            apply_to_range($range, sub {
+                my $point = shift;
+                $point->{'NFG_QC'} = $value;
+            });
+        }
     });
 }
 
@@ -1566,6 +1607,38 @@ sub NamedSequences {
     });
 }
 
+sub tweak_nfg_qc {
+    # See http://www.unicode.org/reports/tr29/tr29-27.html#Grapheme_Cluster_Boundary_Rules
+    for my $point (values %$points_by_code) {
+        my $code = $point->{'code'};
+
+        # \r
+        if ($code == 0x0D) {
+            $point->{'NFG_QC'} = 0;
+        }
+
+        # Hangul
+        elsif ($point->{'Hangul_Syllable_Type'}) {
+            $point->{'NFG_QC'} = 0;
+        }
+
+        # Regional indicators
+        elsif ($code >= 0x1F1E6 && $code <= 0x1F1FF) {
+            $point->{'NFG_QC'} = 0;
+        }
+
+        # Grapheme_Extend
+        elsif ($point->{'Grapheme_Extend'}) {
+            $point->{'NFG_QC'} = 0;
+        }
+
+        # SpacingMark, and a couple of specials
+        elsif ($point->{'gencat_name'} eq 'Mc' || $code == 0x0E33 || $code == 0x0EB3) {
+            $point->{'NFG_QC'} = 0;
+        }
+    }
+}
+
 sub register_binary_property {
     my $name = shift;
     $all_properties->{$name} = $binary_properties->{$name} = {
